//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************


#include <stdio.h>
#include <time.h>

#if defined(WIN32)
  #include <windows.h>
  #include <winsock2.h>
  #include <ws2tcpip.h>
#else
  #include "windowsdef.h"
  #include <stdlib.h>
  #include <string.h>
  #include <sys/types.h>
  #include <sys/socket.h>
  #include <sys/ioctl.h>
  #include <errno.h>
  #include <unistd.h>
  #include <fcntl.h>
  #include <netinet/in.h>
  #include <limits.h>
  #include <netdb.h>
  #include <arpa/inet.h>
#endif

#include "aes.h"
#include "clientProcessPkt.h"
#include "aesCBC.h"
#include "version.h"
#include "randomGen.h"
#include "md5.h"
#include "platformSpecificFunc.h"

extern DWORD gOurIPAddress;
extern unsigned short int clientPortLow;
extern unsigned short int clientPortHigh;
extern unsigned short int clientService;



//************************************************************************************
// Determines the IP address of the interface that the IP packet will be sent out
// from.
//
//************************************************************************************

BOOL getOutInterfaceIP(DWORD destIP, DWORD *outIfIP)
{
// TODO : Check utility of this function
    dk_socket_t sock;

	struct sockaddr_in sdest;

	sdest.sin_family = AF_INET;
	sdest.sin_port = 0;       
	sdest.sin_addr.s_addr = destIP;		// dest addr
	
	if(connect(sock, (struct sockaddr *)&sdest, sizeof(sdest)) == -1)
	{
//		printf("Cannot connect!\n");
		dk_closesocket(sock);
		return FALSE;
	}

	struct sockaddr_in ouraddr;
	socklen_t sinSize = sizeof(ouraddr);
	
	if(getsockname(sock, (struct sockaddr *)&ouraddr, &sinSize) == -1)
	{
//		printf("getsockname error!\n");
		dk_closesocket(sock);
		return FALSE;
	}

	*outIfIP = ouraddr.sin_addr.s_addr;
	dk_closesocket(sock);
	return TRUE;
}

//************************************************************************************
// Returns "size" padded up to 16 byte boundary.
// If size is already 0 mod 16, then return size + 16.
//
//************************************************************************************

int getPaddedSize(int size)
{
	int result = (((size - 1) >> 4) + 1) << 4;		// aligned to 16 bytes
	if(result == size)
		result += 16;

	return result;
}

//************************************************************************************
// Returns a buffer containing data padded up to boundary of 16 bytes and AES
// encrypted using userID and password hash.
//
// This buffer must be freed after use.
//
//************************************************************************************

char *padAndEncrypt(char *userID, unsigned char *userPasswordHash, char *data, int size, int *outSize)
{
	int paddedSize = getPaddedSize(size);
	int reqSize;

	if(paddedSize == size)
		reqSize = size + 16;
	else 
		reqSize = paddedSize;

	char *result = (char *)malloc(reqSize);
	if(result)
	{
		memcpy(result, data, size);
		int padLen = reqSize - size;

		unsigned char *ptr = (unsigned char *)result + size;
		for(int i = size; i < reqSize; i++)
		{
			result[i] = (unsigned char)padLen;
		}

		unsigned char encKey[16];
		genEncKey(userID, userPasswordHash, encKey);

		aes_ctx aesState;
		aes_enc_key(encKey, sizeof(encKey), &aesState);
		aesEncrypt(&aesState, (unsigned char *)result, reqSize);

		*outSize = reqSize;

		memset(&aesState, 0, sizeof(aes_ctx));
		memset(encKey, 0, sizeof(encKey));		// zero out
	}

	return result;

}

//************************************************************************************
// Generates packet P1 containing 3 SYN knock sequences and 3 ISN numbers.
//
//	unsigned char IV[16];
//	unsigned char MD5[16];
//	unsigned char version;	
//	unsigned short int packetType;		// P1
//	unsigned short int reqType;			// single connection/multiple connections
//	uint32_t internelIP;
//	unsigned short int ports[MAX_KNOCKS];
//	uint32_t ISNs[MAX_KNOCKS];
//	uint32_t timeStamp;
//	unsigned short int service;
//	unsigned short int portLow;
//	unsigned short int portHigh;
//
//************************************************************************************
void generateNetOrderP1(PORTSEQ_REQ *host_pq, PORTSEQ_REQ *netorder_pq)
{
	*netorder_pq = *host_pq;			// copy the entrie struct

	netorder_pq->packetType = htons(host_pq->packetType);
	netorder_pq->reqType = htons(host_pq->reqType);
	netorder_pq->timeStamp = htonl(host_pq->timeStamp);
	netorder_pq->service = htons(host_pq->service);
	netorder_pq->portLow = htons(host_pq->portLow);
	netorder_pq->portHigh = htons(host_pq->portHigh);
    int i;
	for(i = 0; i < MAX_KNOCKS; i++)
		netorder_pq->ports[i] = htons(host_pq->ports[i]);

	for( i = 0; i < MAX_KNOCKS; i++)
		netorder_pq->ISNs[i] = htonl(host_pq->ISNs[i]);
}


void generateKnockSequence(PORTSEQ_REQ *pq, PORTSEQ_REQ *netorder_pq, 
						   char *userID, unsigned char *userPasswordHash)
{
	memset(pq, 0, sizeof(PORTSEQ_REQ));
	pq->packetType = P1;
	pq->version = VERSION;
	pq->reqType = SINGLE_CONNECTION;
    pq->service = clientService;
    pq->portLow = clientPortLow;
    pq->portHigh = clientPortHigh;
	// get our first IP address
	pq->internelIP = gOurIPAddress;

	getRandBytes(pq->IV, 16);
    
    int i;
	for(i = 0; i < MAX_KNOCKS; i++)
		pq->ports[i] = getRandWORD();

	for(i = 0; i < MAX_KNOCKS; i++)
		pq->ISNs[i] = getRandDWORD();

	pq->timeStamp = time(NULL);

	generateNetOrderP1(pq, netorder_pq);
	
	// TODO calc MD5 based on the netseq_pq
	// Calc MD5
	MD5Context ctx;
	MD5Init(&ctx);
	MD5Update(&ctx, (unsigned char *)userID, strlen(userID));
	MD5Update(&ctx, (unsigned char *)userPasswordHash, SIZEOF_USERPASSWORD_HASH);
	MD5Update(&ctx, (unsigned char *)netorder_pq, sizeof(PORTSEQ_REQ));

	md5byte digest[16];
	MD5Final(digest, &ctx);

	for(i = 0; i < 16; i++)
		netorder_pq->MD5[i] = digest[i];

	/*
	for(i = 0; i < sizeof(PORTSEQ_REQ); i++)
		printf("%.2X ", (unsigned char)((char *)pq)[i]);
	*/
}


//************************************************************************************
// Returns a buffer containing packet P1 with 9 bytes of userID appended in
// front.
//
// Buffer must be freed after use.
//
//************************************************************************************

char *makePortSeqReqPkt(char *userID, int userIDSize, char *paddedPq, int size, int *outSize)
{
	if(userIDSize > SIZEOF_USERID - 1)
		userIDSize = SIZEOF_USERID - 1;
	if(userIDSize <= 0)
		return NULL;

	char *result = (char *)malloc(size + SIZEOF_USERID);

	if(result)
	{
		memset(result, 0, size + SIZEOF_USERID);
		memcpy(result, userID, userIDSize);
		memcpy(result + SIZEOF_USERID, paddedPq, size);

		*outSize = size + SIZEOF_USERID;
	}

	return result;
}

//************************************************************************************
// Creates packet P2
//
//	unsigned char version;	
//	unsigned short int packetType;		// P2
//	unsigned short int reqType;			// single connection/multiple connections
//	uint32_t internelIP;
//	uint32_t timeStamp;
//
//
//************************************************************************************
void generateNetOrderP2(PORTREQ_QUERY *host_prc, PORTREQ_QUERY *netorder_prc)
{
	*netorder_prc = *host_prc;

	netorder_prc->packetType = htons(host_prc->packetType);
	netorder_prc->reqType = htons(host_prc->reqType);
	netorder_prc->timeStamp = htonl(host_prc->timeStamp);
}


void generateKnockResultQuery(PORTREQ_QUERY *prc, PORTREQ_QUERY *netorder_prc, 
							  char *userID, unsigned char *userPasswordHash, 
							  DWORD prevTimeStamp)
{
	memset(prc, 0, sizeof(PORTREQ_QUERY));
	prc->packetType = P2;

	// get our first IP address
	prc->internelIP = gOurIPAddress;
	prc->version = VERSION;
	prc->reqType = SINGLE_CONNECTION;
	
	getRandBytes(prc->IV, 16);

	prc->timeStamp = prevTimeStamp;

	generateNetOrderP2(prc, netorder_prc);

	// Calc MD5
	MD5Context ctx;
	MD5Init(&ctx);
	MD5Update(&ctx, (unsigned char *)userID, strlen(userID));
	MD5Update(&ctx, (unsigned char *)userPasswordHash, SIZEOF_USERPASSWORD_HASH);
	MD5Update(&ctx, (unsigned char *)netorder_prc, sizeof(PORTREQ_QUERY));

	md5byte digest[16];
	MD5Final(digest, &ctx);

	for(int i = 0; i < 16; i++)
		netorder_prc->MD5[i] = digest[i];

	/*
	printf("\n\n");
	for(i = 0; i < sizeof(PORTREQ_QUERY); i++)
		printf("%.2X ", (unsigned char)((char *)prc)[i]);
	printf("\n");
	*/
}


//************************************************************************************
// Creates packet PK
//
//	unsigned char version;	
//	unsigned short int packetType;		// PK
//	unsigned short int reqType;			// single connection/multiple connections
//	uint32_t internalIP;
//  uint32_t ISN;
//	uint32_t timeStamp;
//
//
//************************************************************************************
void generateNetOrderPK(PORTKNOCK *host_prc, PORTKNOCK *netorder_prc)
{
	*netorder_prc = *host_prc;

	netorder_prc->packetType = htons(host_prc->packetType);
	netorder_prc->reqType = htons(host_prc->reqType);
	netorder_prc->timeStamp = htonl(host_prc->timeStamp);
	netorder_prc->ISN = htonl(host_prc->ISN);
	
}


void generateKnock(PORTKNOCK *prc, PORTKNOCK *netorder_prc, 
							  char *userID, unsigned char *userPasswordHash, 
							  uint32_t ISN,
							  DWORD prevTimeStamp)
{
	memset(prc, 0, sizeof(PORTKNOCK));
	prc->packetType = PK;

	// get our first IP address
	prc->internelIP = gOurIPAddress;
	prc->version = VERSION;
	prc->reqType = SINGLE_CONNECTION;
	prc->ISN=ISN;
//	printf("generateKnock %d\n",ISN);
	getRandBytes(prc->IV, 16);

	prc->timeStamp = prevTimeStamp;

	generateNetOrderPK(prc, netorder_prc);

	// Calc MD5
	MD5Context ctx;
	MD5Init(&ctx);
	MD5Update(&ctx, (unsigned char *)userID, strlen(userID));
	MD5Update(&ctx, (unsigned char *)userPasswordHash, SIZEOF_USERPASSWORD_HASH);
	MD5Update(&ctx, (unsigned char *)netorder_prc, sizeof(PORTKNOCK));

	md5byte digest[16];
	MD5Final(digest, &ctx);

	for(int i = 0; i < 16; i++)
		netorder_prc->MD5[i] = digest[i];

	/*
	printf("\n\n");
	for(i = 0; i < sizeof(PORTREQ_QUERY); i++)
		printf("%.2X ", (unsigned char)((char *)prc)[i]);
	printf("\n");
	*/
}

//************************************************************************************
// calculates ip, tcp, udp checksum
//
//************************************************************************************

unsigned int calculate_checksum(unsigned short* addr, int count)
{
  unsigned int checksum;
     /* Compute Internet Checksum for "count" bytes
    *         beginning at location "addr".
    */
  uint32_t sum = 0;

    while( count > 1 )  {
     /*  This is the inner loop */
     sum += * ((unsigned short*) addr++);
     count -= 2;
   }

     /*  Add left-over byte, if any */
   if( count > 0 )
     sum += * (unsigned char *) addr;

     /*  Fold 32-bit sum to 16 bits */
   while (sum>>16)
     sum = (sum & 0xffff) + (sum >> 16);

   return (unsigned int) ~sum;
}

//************************************************************************************
// Sends SYN knock packet with specific ISN and dest port
//
//************************************************************************************

BOOL sendSYNKnockPacket(SOCKET s_send, DWORD srcIP, DWORD destIP, DWORD ISN, WORD port)
{
	char datagram[4096];
	struct ipheader *iph = (struct ipheader *) datagram;
    struct tcpheader *tcph = (struct tcpheader *) (datagram + sizeof (struct ipheader));

	memset (datagram, 0, 4096);     // zero out the buffer

    // fill in the ip/tcp header values
    iph->ip_hl			= 5;
    iph->ip_v			= 4;
    iph->ip_tos			= 0;
    iph->ip_len			= htons(sizeof (struct ipheader) + sizeof (struct tcpheader));      // no payload
    iph->ip_id			= htons (getRandWORD());		// the value doesn't matter here
    iph->ip_off			= 0;
    iph->ip_ttl			= 255;
    iph->ip_p			= IPPROTO_TCP;
    iph->ip_sum			= 0;								
    iph->ip_src			= srcIP;
    iph->ip_dst			= destIP;
     
	tcph->th_sport		= htons(getRandWORD());
	tcph->th_dport		= htons(port);
	tcph->th_seq		= htonl(ISN);
	tcph->th_ack		= 0;

	tcph->th_headerlen	= 5;
	tcph->th_resv		= 0;
	tcph->th_flags		= TH_SYN;
	tcph->th_win		= htons(16384);
	tcph->th_sum		= 0;												   
	tcph->th_urp = 0;

    iph->ip_sum = calculate_checksum((unsigned short *) datagram, 20);
    //printf("\nIP Checksum is %X\n", iph->ip_sum);

    
	unsigned char tcp_chksum_buf[sizeof(tcpheader) + sizeof(tcp_pseudoheader)];
    struct tcp_pseudoheader *tcp_pseudohdr = (struct tcp_pseudoheader *)tcp_chksum_buf;

    tcp_pseudohdr->ip_src = iph->ip_src;
    tcp_pseudohdr->ip_dst = iph->ip_dst;
    tcp_pseudohdr->resv = 0;
    tcp_pseudohdr->ip_p = IPPROTO_TCP;
    tcp_pseudohdr->tcp_len = htons(20);

	memcpy(tcp_chksum_buf + sizeof(tcp_pseudoheader), tcph, sizeof(tcpheader));

	tcph->th_sum = calculate_checksum ((unsigned short *)tcp_chksum_buf,
										sizeof(tcpheader) +
										sizeof(tcp_pseudoheader));
	//printf("TCP Checksum is %X\n", tcph->th_sum);

    // the sockaddr_in containing the dest. address is used in sendto() to determine the datagrams path

	sockaddr_in sin;
	sin.sin_family = AF_INET;
	sin.sin_port = 0;       
	sin.sin_addr.s_addr = destIP;		// dest addr
	
	// send knock
	if(sendto(s_send, datagram, 40, 0, (struct sockaddr *) &sin, sizeof (sin)) < 0)
	{
		printf ("Error: RAW socket sendto() error\n");
		return FALSE;
	}
	return TRUE;
}


#if !defined(WIN32)
SOCKET gRawSendSock = INVALID_SOCKET;
BOOL createRawSendSocket(void)
{
	//*************************** Setup RAW socket for spoofing ***************************/

	gRawSendSock = socket (AF_INET, SOCK_RAW, IPPROTO_RAW);
	
	if(gRawSendSock == INVALID_SOCKET)
	{
		printf("Error: RAW socket creation failed: %d\n"
			"You may need to run this program as root.\n",
			errno);
		return FALSE;
	}

	
	int bOpt = 1;
	if (setsockopt (gRawSendSock, IPPROTO_IP, IP_HDRINCL, (const char *)&bOpt, sizeof (bOpt)) == SOCKET_ERROR)
	{
		printf ("Error: Cannot set HDRINCL!\n");
		return FALSE;
	}

	return TRUE;
}
#endif



//************************************************************************************
// Creates RAW socket and send SYN knocks to server.
//
//************************************************************************************

BOOL doSYNPortKnock(char *destIP, PORTSEQ_REQ *pq)
{
	DWORD ourIP;
	if(!getOutInterfaceIP(inet_addr(destIP), &ourIP))
	{
		printf("No route to server!\n");
		return FALSE;
	}


	#if defined(WIN32)

	//*************************** Setup RAW socket for spoofing ***************************/
	SOCKET s_send = socket (AF_INET, SOCK_RAW, IPPROTO_RAW);
	if(s_send == INVALID_SOCKET)
	{
		printf("Error: RAW socket creation failed: %d\n"
			   "You may need to run this program as Administrator.\n",
			   WSAGetLastError());
		return FALSE;
	}

	int bOpt = 1;
	if (setsockopt (s_send, IPPROTO_IP, IP_HDRINCL, (const char *)&bOpt, sizeof (bOpt)) == SOCKET_ERROR)
	{
		printf ("Error: Cannot set HDRINCL!\n");
		return FALSE;
	}

	#else

	SOCKET s_send = gRawSendSock;

	#endif





	struct in_addr ina;
	ina.s_addr = ourIP;
	printf("Sending knock sequences with source address %s\n", inet_ntoa(ina));
	printf("Using timestamp value %u\n\n", pq->timeStamp);

	//************************ Send using identified local address ********************************
	for(int knock = 0; knock < MAX_KNOCKS; knock++)
	{
		printf("Knock? (%d - Port = %5u, ISN = %.8X)\n", knock+1, pq->ports[knock], pq->ISNs[knock]);
		
		// send this packet up to NUM_PARANOID_RESENT number of times to better ensure it'll
		// reach the server

		for(int q = 0; q < NUM_PARANOID_RESENT; q++)
		{
			if(!sendSYNKnockPacket(s_send, ourIP, inet_addr(destIP), 
					       pq->ISNs[knock], pq->ports[knock]))
			{
                #if defined(WIN32)
                closesocket(s_send);
                #else
				close(s_send);
                #endif
				return FALSE;
			}
			#if defined(WIN32)
			Sleep(500);
			#else
			usleep(500000);
			#endif
		}
		#if defined(WIN32)
		Sleep(500);
		#else
		usleep(500000);
		#endif
	}

	#if defined(WIN32)
	closesocket(s_send);
	#else
	close(s_send);
	#endif

	return TRUE;
}

//************************************************************************************
// Creates and Send UDP PK knocks to server.
//
//************************************************************************************

BOOL doUDPPortKnock(char *destIP, PORTSEQ_REQ *pq, char *userID, 
					   unsigned char *userPasswordHash)
{
  

	//************************ Send using identified local IP address ****************************
   
	SOCKET sock;
	for(int knock = 0; knock < MAX_KNOCKS; knock++)
	{
		printf("Knock? (%d - Port = %5u, ISN = %.8X)\n", knock+1, pq->ports[knock], pq->ISNs[knock]);
		
		// send this packet up to NUM_PARANOID_RESENT number of times to better ensure it'll
		// reach the server

     
     
        	PORTKNOCK pk,netorder_pk; 

		    generateKnock(&pk, &netorder_pk, userID, userPasswordHash, pq->ISNs[knock],pq->timeStamp);
        	int pktSize;
        	char *pkt = padAndEncrypt(userID, userPasswordHash, (char *)&netorder_pk, sizeof(PORTKNOCK), &pktSize);

        	//printf("padded size = %d\n", pktSize);

	        // add in the userid in front
        	char *sendPkt = makePortSeqReqPkt(userID, strlen(userID), pkt, pktSize, &pktSize);

	        free(pkt);
/*        	printf("IP : %s\n",destIP);
	        printf("Content : %s\n",sendPkt);
	        printf("Size : %d\n",pktSize);
*/
		for(int q = 0; q < NUM_PARANOID_RESENT; q++)
		{
    		sock = getUDPConnection(destIP, pq->ports[knock], 0);
      		if(sock == INVALID_SOCKET) return FALSE;

       		if(send(sock,(char *)sendPkt, pktSize, 0) == SOCKET_ERROR)
       		{
           		dk_closesocket(sock);
     			return FALSE;
        	}
			dk_sleep(500);
       		dk_closesocket(sock);
		}
		dk_sleep(500);
	}
	return TRUE;
}






//************************************************************************************
// Creates a UDP connection
//
//************************************************************************************

SOCKET getUDPConnection(char *destIP, WORD destPort, WORD srcPort)
{
	SOCKET sock;
	//****************************** Send PortSeq Request PKT ****************************/
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if(sock < 0)
	{
		printf("Error: UDP socket creation failed: %d\n", errno);
		return INVALID_SOCKET;
	}
	
	struct sockaddr_in dest;
	dest.sin_family = AF_INET;
	dest.sin_port = htons(destPort);
	dest.sin_addr.s_addr = inet_addr(destIP);

	struct sockaddr_in local;
	local.sin_family = AF_INET;
	local.sin_port = htons(srcPort);
	local.sin_addr.s_addr = htonl(INADDR_ANY);
		
	if(bind(sock, (struct sockaddr *)&local, sizeof(local)) == SOCKET_ERROR)
	{
		printf("Error: UDP socket bind() failed!\n");
		dk_closesocket(sock);
		return INVALID_SOCKET;
	}

	if(connect(sock, (struct sockaddr *)&dest, sizeof(dest)) == SOCKET_ERROR)
	{
		printf("Error: UDP connect() failed!\n");
		dk_closesocket(sock);
		return INVALID_SOCKET;
	}

	return sock;
}

//************************************************************************************
// Encrypt and sends packet P1
//
//************************************************************************************

BOOL doUDPPortKnockReq(char *destIP, WORD destPort, PORTSEQ_REQ *pq, char *userID, 
					   unsigned char *userPasswordHash)
{
	SOCKET sock = getUDPConnection(destIP, destPort, 0);
	if(sock == INVALID_SOCKET)
		return FALSE;

	int pktSize;
	char *pkt = padAndEncrypt(userID, userPasswordHash, (char *)pq, sizeof(PORTSEQ_REQ), &pktSize);

	//printf("padded size = %d\n", pktSize);

	// add in the userid in front
	char *sendPkt = makePortSeqReqPkt(userID, strlen(userID), pkt, pktSize, &pktSize);

	free(pkt);

	//printf("\nsizeof pkt = %d\n", sizeof(PORTSEQ_REQ));
	

	if(send(sock, (char *)sendPkt, pktSize, 0) != SOCKET_ERROR)
	{
		for(int q = 1; q < NUM_PARANOID_RESENT; q++)
		{
			dk_sleep(500);
			send(sock, (char *)sendPkt, pktSize, 0);		// send again just to be paranoid
		}

		free(sendPkt);
		//printf("Sent\n");
	}
	else
	{
		free(sendPkt);

		printf("Error: UDP send() failed!\n");
		dk_closesocket(sock);
		return FALSE;
	}

	dk_closesocket(sock);
	return TRUE;
}

//************************************************************************************
// Decrypts P3 packet from server
//
//************************************************************************************

BOOL decryptUDPReply(char *udpData, int size, char *userID, unsigned char *userPasswordHash, 
		 uint32_t timeStamp, WORD *outAllocPort, WORD *outPortOpenTime, WORD *outReqType)
{
	if(size % 16)
		return FALSE;

	// AES decrypt code
	PORTREQ_QUERY_REPLY *prqr = (PORTREQ_QUERY_REPLY *)(udpData);	

	unsigned char encKey[16];
	genEncKey(userID, userPasswordHash, encKey);

	aes_ctx aesState;
	aes_dec_key(encKey, sizeof(encKey), &aesState);
	aesDecrypt(&aesState, (unsigned char *)prqr, size);

	if(prqr->version != VERSION)		// 1 byte
		return FALSE;

	// TODO: Do net to host order conversion
	

	// packet type check
	if(ntohs(prqr->packetType) != P3)
	{
		//printf("pkt type error (%d)\n", prqr->packetType);
		return FALSE;
	}

	// MD5 check
	md5byte recvHash[16], digest[16];

	for(int i = 0; i < 16; i++)
	{
		recvHash[i] = prqr->MD5[i];
		prqr->MD5[i] = 0;
	}

	MD5Context ctx;
	MD5Init(&ctx);
	MD5Update(&ctx, (unsigned char *)userID, strlen(userID));
	MD5Update(&ctx, (unsigned char *)userPasswordHash, SIZEOF_USERPASSWORD_HASH);
	MD5Update(&ctx, (unsigned char *)prqr, sizeof(PORTREQ_QUERY_REPLY));

	MD5Final(digest, &ctx);
	if(memcmp(digest, recvHash, 16) != 0)
	{
		//printf("Hash Error\n");
		return FALSE;
	}

	if(ntohl(prqr->timeStamp) != timeStamp)
	{
		//printf("timestamp error (recv %d, %d)\n", prqr->timeStamp, timeStamp);
		return FALSE;
	}

	*outAllocPort = ntohs(prqr->allocatedPort);
	*outPortOpenTime = ntohs(prqr->portOpenTime);
	*outReqType = ntohs(prqr->reqType);
	return TRUE;
}



//************************************************************************************
// Encrypt and sends packet P2
//
//************************************************************************************

BOOL doUDPPortKnockQuery(char *destIP, WORD destPort, PORTREQ_QUERY *prq, PORTREQ_QUERY *netorder_prq,
						 char *userID, unsigned char *userPasswordHash, 
						 WORD *outAllocPort, WORD *outPortTime, WORD *outReqType)
{
	SOCKET sock = getUDPConnection(destIP, destPort, 0);
	if(sock == INVALID_SOCKET)
		return FALSE;
	
	int pktSize;
	char *pkt = padAndEncrypt(userID, userPasswordHash, (char *)netorder_prq, 
		                      sizeof(PORTREQ_QUERY), &pktSize);

	//printf("padded size = %d\n", pktSize);

	// add in the userid in front
	char *sendPkt = makePortSeqReqPkt(userID, strlen(userID), pkt, pktSize, &pktSize);

	free(pkt);

	//printf("\nsizeof pkt = %d\n", sizeof(PORTSEQ_REQ));

	if(send(sock, (char *)sendPkt, pktSize, 0) != SOCKET_ERROR)
	{
		printf("\nCan I come in?  Waiting for response.\n");
		char buffer[1024];
		int recvSize;

		// if server is not firewalled, it'll will send us "dest unreachable" ICMP.
		// so will try till we get a reply
		int triesCount = 0;

		while((recvSize = recv(sock, buffer, sizeof(buffer), 0)) != 
					getPaddedSize(sizeof(PORTREQ_QUERY_REPLY)))
		{
			triesCount++;

			if(triesCount >= 10)
				break;
			dk_sleep(500);

			send(sock, (char *)sendPkt, pktSize, 0);		// resend for 5 times
		}

		free(sendPkt);

		if(recvSize == getPaddedSize(sizeof(PORTREQ_QUERY_REPLY)))
		{
			/*
			printf("Recved %d bytes\n", recvSize);
			for(int i = 0; i < recvSize; i++)
			{
				printf("%.2X ", (unsigned char)buffer[i]);
			}
			*/

			WORD allocPort, openTime, reqType;
			if(decryptUDPReply(buffer, recvSize, userID, userPasswordHash, prq->timeStamp, 
				               &allocPort, &openTime, &reqType))
			{				
				*outAllocPort = allocPort;
				*outPortTime = openTime;
				*outReqType = reqType;
			}			
			else
			{
				dk_closesocket(sock);
				return FALSE;
			}
		}
		else
		{
			dk_closesocket(sock);
			return FALSE;
		}
	}
	else
	{
		free(sendPkt);

		printf("Error: UDP send() failed!\n");
		dk_closesocket(sock);
		return FALSE;
	}

	dk_closesocket(sock);
	return TRUE;
}



