//************************************************************************************
// Copyright (c) 2006, Olivier ROCH VILATO, Chew Keong TAN
// All rights reserved.
//
// Win32/Linux Coding by Chew Keong TAN and Olivier ROCH VILATO
//
// This file is part of DrKnock (based on sig2knock).

// DrKnock is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// DrKnock is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with DrKnock; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//************************************************************************************

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <time.h>
#include "userRecord.h"
#include "portSeqReqList.h"
#include "serverProcessPkt.h"
#include "portForward.h"
#include "randomGen.h"
#include "serverLog.h"
#include "serverConfig.h"
#include "version.h"

#define STALE_TIMEOUT		20

struct in_addr			gOurIPAddr;
char				gOurIPAddrStr[32];
LINKED_LIST			gPortSeqRequestList;
short int           listenAllPorts=0;  // Flag to listen to all port when receiving knocks
short int           connectLock=0;
uint32_t            lockTimeStamp;
short int           clientPortLow=0;
short int           clientPortHigh=0;
#if defined(WIN32)
  #pragma pack(1)
  #include <winsock2.h>
  #include <windows.h>
  #include <ws2tcpip.h> 
  #include <pcap.h>

  #define __GNU_LIBRARY__
  #include "getopt.h"
  #include "win32Service.h"

  #pragma comment(lib,"ws2_32.lib")
  #pragma comment(lib, "wpcap.lib")

  CRITICAL_SECTION	gPortReqRequestListControl;
  extern char		*optarg;
  extern int		optind, optopt;
  SERVICE_STATUS_HANDLE	gSvcHandle;
  bool			gStopFlag = false;
  pcap_t		*gPcapHandle = NULL;
#else
  #include <stdlib.h>
  #include <semaphore.h>
  #include <sys/types.h>
  #include <sys/stat.h>
  #include <fcntl.h>
  #include <unistd.h>
  #include <sys/poll.h>
  #include <pthread.h>
  #include <net/if.h>
  #include <signal.h>
  #include <pwd.h>
  #include <sys/wait.h>
  #include <dirent.h>

  #include "pcap.h"
  #include "windowsdef.h"
  #include "firewallSupport.h"
  #include "privProcSupport.h"
  BOOL				gTerminateApp = FALSE;
  pcap_t          		*adhandle = NULL;
  BOOL            		gDaemon = FALSE;
  BOOL				gUseSyslog = FALSE; 	// syslog support is broken
  pid_t       			gPktProcPid = 0;
  uid_t				gDropId;
  gid_t				gDropGrp;

#endif




//************************************************************************************
// Callback function to remove client port sequence request records from list.
// When a new port knock request is received, old ones will be removed, so each user
// can only have one outstanding knock request.
//
// This is to prevent memory DoS on the server
//
//************************************************************************************

BOOL deletePrevPortSeqRecFunc(LPVOID data, LPVOID toFind)
{
	PORTSEQ_REQREC *aData = (PORTSEQ_REQREC *)data;
	PORTSEQ_REQREC *aFind = (PORTSEQ_REQREC *)toFind;

	if(strcmp(aData->userID, aFind->userID) == 0 &&
	   aData->timeStamp < aFind->timeStamp)		// let user have only one req at a time
	{
		free(aData);
		return TRUE;
	}
	else
		return FALSE;
}

//************************************************************************************
// Callback function that returns TRUE if a port sequence request record packet with
// the same user name and timestamp is already present.
//
// This ensures a new record is not created for duplicated or replyed requests.
//
//************************************************************************************


BOOL hasSameTimeStampPortSeqRec(LPVOID data, LPVOID toFind)
{
	PORTSEQ_REQREC *aData = (PORTSEQ_REQREC *)data;
	PORTSEQ_REQREC *aFind = (PORTSEQ_REQREC *)toFind;

	if(strcmp(aData->userID, aFind->userID) == 0 &&
	   aData->timeStamp == aFind->timeStamp)		// let user have only one req at a time		
		return TRUE;
	else
		return FALSE;
}

//************************************************************************************
// This callback function will delete port sequence request records that are >= 20
// seconds old.  i.e. all port knocks must complete within 20 seconds.
//
// toFind == NULL
//************************************************************************************

BOOL deleteStalePortSeqRecFunc(LPVOID data, LPVOID toFind)
{
        PORTSEQ_REQREC *aData = (PORTSEQ_REQREC *)data;

        if((time(NULL) - aData->insertTime) > STALE_TIMEOUT)       // delete requests after 20 sec
        {
                printf("Delete record %s\n", aData->userID);
                free(aData);
                return TRUE;                            // return TRUE will delete this record
        }
        else
                return FALSE;
}


//************************************************************************************
// Adds a new port sequence request to list.
//
//************************************************************************************

void addPortSeqRequestToList(PORTSEQ_REQ *portReq, DWORD clientIP, char *userID)
{
	PORTSEQ_REQREC *newRec = (PORTSEQ_REQREC *)malloc(sizeof(PORTSEQ_REQREC));

	if(newRec)
	{
		memset(newRec, 0, sizeof(PORTSEQ_REQREC));		// clear mem

		newRec->clientInternalIP = portReq->internelIP;
		newRec->clientIP = clientIP;
        int i;
		for(i = 0; i < MAX_KNOCKS; i++)
			newRec->ISNs[i] = portReq->ISNs[i];
		
		for(i = 0; i < MAX_KNOCKS; i++)
			newRec->ports[i] = portReq->ports[i];

		newRec->nextKnock = 0;
		newRec->timeStamp = portReq->timeStamp;
		strncpy(newRec->userID, userID, sizeof(newRec->userID));

		newRec->userID[sizeof(newRec->userID) - 1] = 0;
		newRec->portAllocated = 0;
		newRec->insertTime = time(NULL);
		newRec->reqType = portReq->reqType;

		#if defined(WIN32)
		EnterCriticalSection(&gPortReqRequestListControl);
		#endif
		// delete user's previous port seq requests.  to prevent malicious user from
		// DoS-ing server
		deleteMultipleItems(&gPortSeqRequestList, deletePrevPortSeqRecFunc, newRec);

		// delete stale records
		deleteMultipleItems(&gPortSeqRequestList, deleteStalePortSeqRecFunc, NULL);

		if(findItem(&gPortSeqRequestList, hasSameTimeStampPortSeqRec, newRec))
		{
			//printf("Previous SAME rec found!\n");
			free(newRec);		// replay, or duplicated packet
		}
		else
		{
			if(!insertItem(&gPortSeqRequestList, (LPVOID)newRec))
				free(newRec);
			else
			{
				updateUserTimestamp(USER_FILENAME, userID, newRec->timeStamp);
				
				struct in_addr ina1, ina2;
			#if defined(WIN32)
				ina1.S_un.S_addr = newRec->clientIP;
				ina2.S_un.S_addr = newRec->clientInternalIP;
			#else
				ina1.s_addr = newRec->clientIP;
				ina2.s_addr = newRec->clientInternalIP;
			#endif
				char ipAddrStr[30];
				memset(ipAddrStr, 0, sizeof(ipAddrStr));
				strncpy(ipAddrStr, inet_ntoa(ina1), sizeof(ipAddrStr) - 1);

				logString("Port request received from %s. (%s, %s)\n", 
					      newRec->userID, inet_ntoa(ina2), ipAddrStr);
			}
		}

		#if defined(WIN32)
		LeaveCriticalSection(&gPortReqRequestListControl);
		#endif
	}
}


//************************************************************************************
// Prints all port sequence request records in list.
//
//************************************************************************************

void printPortSeqReqRec(LPVOID data)
{
	PORTSEQ_REQREC *pq = (PORTSEQ_REQREC *)data;

	struct in_addr inaddr;

  #if defined(WIN32)
	inaddr.S_un.S_addr = pq->clientIP;
	printf("Client Ext IP = %s\n", inet_ntoa(inaddr));
	inaddr.S_un.S_addr = pq->clientInternalIP;
	printf("Client Int IP = %s\n", inet_ntoa(inaddr));
	printf("Client UserID = %s\n", pq->userID);
	printf("Insert Time = %d\n", pq->insertTime);
	printf("%d\n", pq->ISNs[0]);
  #else
	inaddr.s_addr = pq->clientIP;
	printf("Client Ext IP = %s\n", inet_ntoa(inaddr));

	inaddr.s_addr = pq->clientInternalIP;
	printf("Client Int IP = %s\n", inet_ntoa(inaddr));

	printf("Client UserID = %s\n", pq->userID);
	printf("Insert Time = %d\n", pq->insertTime);

	printf("%d\n", pq->ISNs[0]);
  #endif
}


//************************************************************************************
// Process packet P2 received from client.  Sends reply packet P3 if request is valid.
//
//************************************************************************************

BOOL processUDPKnockQuery(LPVOID portSeqReqRec, LPVOID portReqQuery)
{
	PORTSEQ_REQREC *pq = (PORTSEQ_REQREC *)portSeqReqRec;

	PORTREQ_QUERY_REC *pqr = (PORTREQ_QUERY_REC *)portReqQuery;
	PORTREQ_QUERY *prq = &(pqr->prq);
	char *userid = pqr->userID;
	DWORD srcIP = pqr->clientIP;


	if(pq->clientIP == srcIP &&
	   pq->clientInternalIP == prq->internelIP &&	   
	   pq->timeStamp == prq->timeStamp &&
	   pq->nextKnock == MAX_KNOCKS &&
	   pq->reqType == prq->reqType &&
	   strcmp(pq->userID, userid) == 0 &&
	   time(NULL) - pq->insertTime <= STALE_TIMEOUT)
	{
		//printf("Found Req\n");
		//printf("userid = %s, ip = %X, srcPort = %d\n", userid, srcIP, pqr->clientPort);

		USER_RECORD userRec;
		
		#if defined(WIN32)
			if(!getUserRecord(USER_FILENAME, pq->userID, &userRec))
		#else
			if(!psGetUserRecord(pq->userID, &userRec))
		#endif
		{
			logString("Error: getUserRecord() failed in ProcessUDPKnockQuery\n");
			return FALSE;
		}

		if(pq->portAllocated == 0)
		{
			WORD portNum;
			SOCKET s = bindFreePort(srcIP, &portNum);
			if(s != INVALID_SOCKET)
			{
				struct in_addr ina1, ina2,ina3;
	 		  #if defined(WIN32)
 				ina1.S_un.S_addr = pq->clientIP;
				ina2.S_un.S_addr = pq->clientInternalIP;
			  #else
 				ina1.s_addr = pq->clientIP;
				ina2.s_addr = pq->clientInternalIP;
			  #endif
				char ipAddrStr[30];
				memset(ipAddrStr, 0, sizeof(ipAddrStr));
				strncpy(ipAddrStr, inet_ntoa(ina1), sizeof(ipAddrStr) - 1);

				logString("Binded %s to port %u for %s. (%s, %s)\n",getConfigCurrentSection(), 
					      portNum, pq->userID, inet_ntoa(ina2), ipAddrStr);
				
				DWORD threadID;
				FORWARDER_DATA *fdata;

				fdata = (FORWARDER_DATA *)malloc(sizeof(FORWARDER_DATA));
				if(fdata)
				{
					fdata->s = s;
					fdata->allowedIP = srcIP;
					fdata->portNum = portNum;
					pq->portAllocated = portNum;
					pq->replyCount = 1;

					sendPortSeqQueryReply(srcIP, pqr->clientPort, pq->timeStamp, pq->userID,
						                  userRec.userPasswordHash, portNum, pq->reqType);
					#if defined(WIN32)
					  CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)forwarderThreadFunc, (LPVOID)fdata, 0, &threadID);
					#else
    					  pthread_create(&(fdata->pt), NULL, forwarderThreadFunc, (LPVOID)fdata);
					#endif
				}
			}
		}
		else
		{
			if(pq->replyCount < 5)	// limit the number of times we'll reply to the client
			{
				sendPortSeqQueryReply(srcIP, pqr->clientPort, pq->timeStamp, pq->userID,
				                 userRec.userPasswordHash, pq->portAllocated, pq->reqType);
				pq->replyCount++;
			}
		}
		connectLock=0;
		return TRUE;
	}

	return FALSE;
}

typedef struct _SYN_KNOCK
{
	DWORD clientIP;
	DWORD ISN;
	WORD port;
} SYN_KNOCK;

//************************************************************************************
// Process TCP SYN knock packets received.
//
//************************************************************************************

BOOL processSYNKnock(LPVOID portSeqReqRec, LPVOID synKnock)
{
	SYN_KNOCK *sk = (SYN_KNOCK *)synKnock;
	PORTSEQ_REQREC *pq = (PORTSEQ_REQREC *)portSeqReqRec;

	if(pq->nextKnock < MAX_KNOCKS && pq->portAllocated == 0)
	{		
		if(pq->ISNs[pq->nextKnock] == sk->ISN &&
		   pq->ports[pq->nextKnock] == sk->port &&
		   pq->clientIP == sk->clientIP &&
		   time(NULL) - pq->insertTime <= STALE_TIMEOUT)
		{
			pq->nextKnock++;
			
			// printf("Match.\n");

			if(pq->nextKnock == MAX_KNOCKS)
			{
				struct in_addr ina1, ina2;
	 		  #if defined(WIN32)
 				ina1.S_un.S_addr = pq->clientIP;
				ina2.S_un.S_addr = pq->clientInternalIP;
			  #else
 				ina1.s_addr = pq->clientIP;
				ina2.s_addr = pq->clientInternalIP;
			  #endif

				char ipAddrStr[30];
				memset(ipAddrStr, 0, sizeof(ipAddrStr));
				strncpy(ipAddrStr, inet_ntoa(ina1), sizeof(ipAddrStr) - 1);

				logString("Correct knock sequence received from %s (%s, %s).\n", 
						pq->userID, inet_ntoa(ina2), ipAddrStr);
			}
			return TRUE;
		}
	}
	return FALSE;
}

//************************************************************************************
// Callback function invoked by libpcap for every incoming packet.
//
//
//************************************************************************************

void packet_handler(u_char *param, const struct pcap_pkthdr *header, const u_char *pkt_data)
{
     #if defined(WIN32)
	if(pkt_data && !gStopFlag)
     #else
	if(pkt_data && header)
     #endif
	{
		//printf("Recv %d\n", header->caplen);
		IP_HEADER *iphdr;
		UDP_HEADER *udphdr;
		int outDataSize;

		int IPPktLen;
		char *buffer = (char *)getIPHeader((char *)pkt_data, header->caplen, &IPPktLen);   

		if(buffer)
		{
			DWORD ipPktType;

			//printf("IP Pktlen = %d\n", IPPktLen);
			if((ipPktType = getIPPktType(buffer, IPPktLen)) == IPPROTO_UDP)
			{
				char *dataPtr = (char *)getUDPData(buffer, IPPktLen, &gOurIPAddr, 
					                               &iphdr, &udphdr, &outDataSize);
                
				if(dataPtr)
				{					
					char *userid;
					LPVOID decPkt;
					unsigned short pktType;

					if(decryptUDPReq(dataPtr, outDataSize, &userid, (LPVOID *)&decPkt, &pktType))
					{	
                        
                        					
						if((pktType == P1) &&((!connectLock)||((time(NULL)-lockTimeStamp)>15)))
						{
                            connectLock=1;
                            lockTimeStamp=time(NULL);
							PORTSEQ_REQ *portReq = (PORTSEQ_REQ *)decPkt;
                            logString("Service request : %d\n",portReq->service);
                            logString("Port Request : %d-%d\n",portReq->portLow,portReq->portHigh);
                            if (getConfigClientCanSetPort(userid))
                            {
                              clientPortLow=portReq->portLow;
                              clientPortHigh=portReq->portHigh;
                            }
                            else
                            {
                              clientPortLow=0;
                              clientPortHigh=0;
                            }
                            // Fix It : Maybe set to early
                            if (portReq->service) setUserService(userid,portReq->service);							


							// critical section will be aquired within addPortSeqRequestToList func
							addPortSeqRequestToList(portReq, getIPSrcAddr(buffer, IPPktLen), userid);
                            // prepare to listen to all ports
                            listenAllPorts=1;
						}
						else if (pktType == P2) // packType == P2
						{
							PORTREQ_QUERY_REC pqr;
							memset(&pqr, 0, sizeof(PORTREQ_QUERY_REC));

							pqr.prq = *((PORTREQ_QUERY *)decPkt);
							strncpy(pqr.userID, userid, SIZEOF_USERID);
							pqr.clientIP = getIPSrcAddr(buffer, IPPktLen);
							pqr.clientPort = getUDPSrcPort(buffer, IPPktLen);
							
							#if defined(WIN32)
							  EnterCriticalSection(&gPortReqRequestListControl);
							#endif
							// delete stale records
							deleteMultipleItems(&gPortSeqRequestList,
												deleteStalePortSeqRecFunc, NULL);

							if(processSingleItem(&gPortSeqRequestList, 
												 processUDPKnockQuery, 
												 &pqr))
							{
							}
							#if defined(WIN32)
							  LeaveCriticalSection(&gPortReqRequestListControl);
							#endif
						}
						else if(pktType == PK)  // Knock
      					{
		    				//printf("Knock!!\n");
	     					PORTKNOCK *portknock=(PORTKNOCK *) decPkt;
			    			SYN_KNOCK synKnock;
          					synKnock.clientIP=iphdr->ip_src;
	               			synKnock.ISN = portknock->ISN;
        	           		synKnock.port=ntohs(udphdr->th_dport);

						#if defined(WIN32)
             		  	    EnterCriticalSection(&gPortReqRequestListControl);
						#endif

                      		processSingleItem(&gPortSeqRequestList,processSYNKnock,&synKnock);

						#if defined(WIN32)
						    LeaveCriticalSection(&gPortReqRequestListControl);
						#endif
                        }
					}

				}
			}
			else if(ipPktType == IPPROTO_TCP)
			{
				DWORD ISN;
				WORD port;
				struct in_addr ipAddr;

				if(getTCPKnockData(buffer, IPPktLen, &gOurIPAddr, &ISN, &port, &ipAddr))
				{						
					//printf("TCP knock port = %d, ISN = %X, %s\n", port, ISN, inet_ntoa(ipAddr));
					SYN_KNOCK synKnock;
					#if defined(WIN32)
  					  synKnock.clientIP = ipAddr.S_un.S_addr;
					#else
  					  synKnock.clientIP = ipAddr.s_addr;
					#endif
					synKnock.ISN = ISN;
					synKnock.port = port;

					#if defined(WIN32)
					  EnterCriticalSection(&gPortReqRequestListControl);
					#endif

					processSingleItem(&gPortSeqRequestList, processSYNKnock, &synKnock);

					#if defined(WIN32)
					  LeaveCriticalSection(&gPortReqRequestListControl);
					#endif
				}
			}
		}
	}
}


#if !defined(WIN32)
  #include "unix/drknockd_unix.cpp"
#else
  #include "win32/drknockd_win32.cpp"
#endif

